@@ -1,45 +1,28 @@
 import os
-import sys
 from collections import OrderedDict
 from copy import deepcopy
 
+from ..utils.constants import DEFAULT_CONFIG_FILE
 from ..utils.fs import locate_file
 from .utils import get_normalized_dependency, is_valid_project_name, normalize_project_name
 
-# TODO: remove when we drop Python 2
-if sys.version_info[0] < 3:  # no cov
-    from io import open
-
-    import toml
-
-    from ..utils.compat import byteify_object
-
-    def load_toml(path):
-        with open(path, 'r', encoding='utf-8') as f:
-            # this is to support any `isinstance(metadata_entry, str)`
-            return byteify_object(toml.loads(f.read()))
-
-    def ensure_string(obj):
-        if not isinstance(obj, bytes):
-            return obj.encode('utf-8')
-        return obj
-
-else:
-    import tomli
-
-    def load_toml(path):
-        with open(path, 'r', encoding='utf-8') as f:
-            return tomli.loads(f.read())
-
-    def ensure_string(obj):
-        return obj
+try:
+    import tomllib
+except ImportError:
+    import tomli as tomllib  # type: ignore
 
 
-class ProjectMetadata(object):
+def load_toml(path):
+    with open(path, 'r', encoding='utf-8') as f:
+        return tomllib.loads(f.read())
+
+
+class ProjectMetadata:
     def __init__(self, root, plugin_manager, config=None):
         self.root = root
         self.plugin_manager = plugin_manager
         self._config = config
+        self._context = None
         self._build = None
         self._core = None
         self._hatch = None
@@ -51,6 +34,15 @@ class ProjectMetadata(object):
         if config is not None and root is not None:
             self._project_file = os.path.join(root, 'pyproject.toml')
 
+    @property
+    def context(self):
+        if self._context is None:
+            from ..utils.context import Context
+
+            self._context = Context(self.root)
+
+        return self._context
+
     @property
     def version(self):
         """
@@ -111,8 +103,8 @@ class ProjectMetadata(object):
                         metadata.dynamic.remove(new_field)
                     else:
                         raise ValueError(
-                            'The field `{}` was set dynamically and therefore must '
-                            'be listed in `project.dynamic`'.format(new_field)
+                            f'The field `{new_field}` was set dynamically and therefore must be '
+                            f'listed in `project.dynamic`'
                         )
 
             self._core = metadata
@@ -131,9 +123,9 @@ class ProjectMetadata(object):
                 raise TypeError('The `tool.hatch` configuration must be a table')
 
             if self._project_file is not None:
-                hatch_file = os.path.join(os.path.dirname(self._project_file), 'hatch.toml')
+                hatch_file = os.path.join(os.path.dirname(self._project_file), DEFAULT_CONFIG_FILE)
             else:
-                hatch_file = locate_file(self.root, 'hatch.toml')
+                hatch_file = locate_file(self.root, DEFAULT_CONFIG_FILE)
 
             if hatch_file and os.path.isfile(hatch_file):
                 config = load_toml(hatch_file)
@@ -150,14 +142,29 @@ class ProjectMetadata(object):
 
         version = core_metadata.version
         if version is None:
-            version = self.hatch.version
+            version = self.hatch.version.cached
+            source = f'source `{self.hatch.version.source_name}`'
+        else:
+            source = 'field `project.version`'
 
-        from packaging.version import Version
+        from packaging.version import InvalidVersion, Version
 
-        self._version = str(Version(version))
+        try:
+            normalized_version = str(Version(version))
+        except InvalidVersion:
+            raise ValueError(
+                # Put text on its own line to prevent mostly duplicate output
+                f'Invalid version `{version}` from {source}, see https://peps.python.org/pep-0440/'
+            )
+        else:
+            self._version = normalized_version
+
+    def validate_fields(self):
+        _ = self.version
+        self.core.validate_fields()
 
 
-class BuildMetadata(object):
+class BuildMetadata:
     """
     https://peps.python.org/pep-0517/
     """
@@ -184,12 +191,12 @@ class BuildMetadata(object):
 
             for i, entry in enumerate(requires, 1):
                 if not isinstance(entry, str):
-                    raise TypeError('Dependency #{} of field `build-system.requires` must be a string'.format(i))
+                    raise TypeError(f'Dependency #{i} of field `build-system.requires` must be a string')
 
                 try:
                     requires_complex.append(Requirement(entry))
                 except InvalidRequirement as e:
-                    raise ValueError('Dependency #{} of field `build-system.requires` is invalid: {}'.format(i, e))
+                    raise ValueError(f'Dependency #{i} of field `build-system.requires` is invalid: {e}')
 
             self._requires_complex = requires_complex
 
@@ -222,14 +229,14 @@ class BuildMetadata(object):
 
             for i, entry in enumerate(backend_path, 1):
                 if not isinstance(entry, str):
-                    raise TypeError('Entry #{} of field `build-system.backend-path` must be a string'.format(i))
+                    raise TypeError(f'Entry #{i} of field `build-system.backend-path` must be a string')
 
             self._backend_path = backend_path
 
         return self._backend_path
 
 
-class CoreMetadata(object):
+class CoreMetadata:
     """
     https://peps.python.org/pep-0621/
     """
@@ -373,12 +380,12 @@ class CoreMetadata(object):
                     content_type = 'text/plain'
                 else:
                     raise TypeError(
-                        'Unable to determine the content-type based on the extension of readme file: {}'.format(readme)
+                        f'Unable to determine the content-type based on the extension of readme file: {readme}'
                     )
 
                 readme_path = os.path.normpath(os.path.join(self.root, readme))
                 if not os.path.isfile(readme_path):
-                    raise OSError('Readme file does not exist: {}'.format(readme))
+                    raise OSError(f'Readme file does not exist: {readme}')
 
                 with open(readme_path, 'r', encoding='utf-8') as f:
                     self._readme = f.read()
@@ -407,7 +414,7 @@ class CoreMetadata(object):
 
                     path = os.path.normpath(os.path.join(self.root, relative_path))
                     if not os.path.isfile(path):
-                        raise OSError('Readme file does not exist: {}'.format(relative_path))
+                        raise OSError(f'Readme file does not exist: {relative_path}')
 
                     with open(path, 'r', encoding=readme.get('charset', 'utf-8')) as f:
                         contents = f.read()
@@ -428,7 +435,7 @@ class CoreMetadata(object):
             else:
                 raise TypeError('Field `project.readme` must be a string or a table')
 
-            self._readme = ensure_string(self._readme)
+            self._readme = self._readme
 
         return self._readme
 
@@ -476,7 +483,7 @@ class CoreMetadata(object):
             try:
                 self._python_constraint = SpecifierSet(requires_python)
             except InvalidSpecifier as e:
-                raise ValueError('Field `project.requires-python` is invalid: {}'.format(e))
+                raise ValueError(f'Field `project.requires-python` is invalid: {e}')
 
             self._requires_python = str(self._python_constraint)
 
@@ -524,7 +531,7 @@ class CoreMetadata(object):
 
                     path = os.path.normpath(os.path.join(self.root, relative_path))
                     if not os.path.isfile(path):
-                        raise OSError('License file does not exist: {}'.format(relative_path))
+                        raise OSError(f'License file does not exist: {relative_path}')
 
                     with open(path, 'r', encoding='utf-8') as f:
                         contents = f.read()
@@ -582,12 +589,12 @@ class CoreMetadata(object):
                 for i, relative_path in enumerate(paths, 1):
                     if not isinstance(relative_path, str):
                         raise TypeError(
-                            'Entry #{} in field `paths` in the `project.license-files` table must be a string'.format(i)
+                            f'Entry #{i} in field `paths` in the `project.license-files` table must be a string'
                         )
 
                     path = os.path.normpath(os.path.join(self.root, relative_path))
                     if not os.path.isfile(path):
-                        raise OSError('License file does not exist: {}'.format(relative_path))
+                        raise OSError(f'License file does not exist: {relative_path}')
 
                     license_files.append(os.path.relpath(path, self.root).replace('\\', '/'))
             elif 'globs' in data:
@@ -600,7 +607,7 @@ class CoreMetadata(object):
                 for i, pattern in enumerate(globs, 1):
                     if not isinstance(pattern, str):
                         raise TypeError(
-                            'Entry #{} in field `globs` in the `project.license-files` table must be a string'.format(i)
+                            f'Entry #{i} in field `globs` in the `project.license-files` table must be a string'
                         )
 
                     full_pattern = os.path.normpath(os.path.join(self.root, pattern))
@@ -635,30 +642,22 @@ class CoreMetadata(object):
             if not isinstance(authors, list):
                 raise TypeError('Field `project.authors` must be an array')
 
-            try:
-                from email.headerregistry import Address
-            # TODO: remove when we drop Python 2
-            except ImportError:  # no cov
-                Address = (
-                    lambda display_name='', addr_spec='': addr_spec
-                    if not display_name
-                    else '{} <{}>'.format(display_name, addr_spec)
-                )
+            from email.headerregistry import Address
 
             authors = deepcopy(authors)
             authors_data = {'name': [], 'email': []}
 
             for i, data in enumerate(authors, 1):
                 if not isinstance(data, dict):
-                    raise TypeError('Author #{} of field `project.authors` must be an inline table'.format(i))
+                    raise TypeError(f'Author #{i} of field `project.authors` must be an inline table')
 
                 name = data.get('name', '')
                 if not isinstance(name, str):
-                    raise TypeError('Name of author #{} of field `project.authors` must be a string'.format(i))
+                    raise TypeError(f'Name of author #{i} of field `project.authors` must be a string')
 
                 email = data.get('email', '')
                 if not isinstance(email, str):
-                    raise TypeError('Email of author #{} of field `project.authors` must be a string'.format(i))
+                    raise TypeError(f'Email of author #{i} of field `project.authors` must be a string')
 
                 if name and email:
                     authors_data['email'].append(str(Address(display_name=name, addr_spec=email)))
@@ -667,9 +666,7 @@ class CoreMetadata(object):
                 elif name:
                     authors_data['name'].append(name)
                 else:
-                    raise ValueError(
-                        'Author #{} of field `project.authors` must specify either `name` or `email`'.format(i)
-                    )
+                    raise ValueError(f'Author #{i} of field `project.authors` must specify either `name` or `email`')
 
             self._authors = authors
             self._authors_data = authors_data
@@ -705,30 +702,22 @@ class CoreMetadata(object):
             if not isinstance(maintainers, list):
                 raise TypeError('Field `project.maintainers` must be an array')
 
-            try:
-                from email.headerregistry import Address
-            # TODO: remove when we drop Python 2
-            except ImportError:  # no cov
-                Address = (
-                    lambda display_name='', addr_spec='': addr_spec
-                    if not display_name
-                    else '{} <{}>'.format(display_name, addr_spec)
-                )
+            from email.headerregistry import Address
 
             maintainers = deepcopy(maintainers)
             maintainers_data = {'name': [], 'email': []}
 
             for i, data in enumerate(maintainers, 1):
                 if not isinstance(data, dict):
-                    raise TypeError('Maintainer #{} of field `project.maintainers` must be an inline table'.format(i))
+                    raise TypeError(f'Maintainer #{i} of field `project.maintainers` must be an inline table')
 
                 name = data.get('name', '')
                 if not isinstance(name, str):
-                    raise TypeError('Name of maintainer #{} of field `project.maintainers` must be a string'.format(i))
+                    raise TypeError(f'Name of maintainer #{i} of field `project.maintainers` must be a string')
 
                 email = data.get('email', '')
                 if not isinstance(email, str):
-                    raise TypeError('Email of maintainer #{} of field `project.maintainers` must be a string'.format(i))
+                    raise TypeError(f'Email of maintainer #{i} of field `project.maintainers` must be a string')
 
                 if name and email:
                     maintainers_data['email'].append(str(Address(display_name=name, addr_spec=email)))
@@ -738,7 +727,7 @@ class CoreMetadata(object):
                     maintainers_data['name'].append(name)
                 else:
                     raise ValueError(
-                        'Maintainer #{} of field `project.maintainers` must specify either `name` or `email`'.format(i)
+                        f'Maintainer #{i} of field `project.maintainers` must specify either `name` or `email`'
                     )
 
             self._maintainers = maintainers
@@ -779,7 +768,7 @@ class CoreMetadata(object):
 
             for i, keyword in enumerate(keywords, 1):
                 if not isinstance(keyword, str):
-                    raise TypeError('Keyword #{} of field `project.keywords` must be a string'.format(i))
+                    raise TypeError(f'Keyword #{i} of field `project.keywords` must be a string')
 
                 unique_keywords.add(keyword)
 
@@ -810,7 +799,7 @@ class CoreMetadata(object):
 
             for i, classifier in enumerate(classifiers, 1):
                 if not isinstance(classifier, str):
-                    raise TypeError('Classifier #{} of field `project.classifiers` must be a string'.format(i))
+                    raise TypeError(f'Classifier #{i} of field `project.classifiers` must be a string')
 
                 unique_classifiers.add(classifier)
 
@@ -840,7 +829,7 @@ class CoreMetadata(object):
 
             for label, url in sorted(urls.items()):
                 if not isinstance(url, str):
-                    raise TypeError('URL `{}` of field `project.urls` must be a string'.format(label))
+                    raise TypeError(f'URL `{label}` of field `project.urls` must be a string')
 
                 sorted_urls[label] = url
 
@@ -871,7 +860,7 @@ class CoreMetadata(object):
 
             for name, object_ref in sorted(scripts.items()):
                 if not isinstance(object_ref, str):
-                    raise TypeError('Object reference `{}` of field `project.scripts` must be a string'.format(name))
+                    raise TypeError(f'Object reference `{name}` of field `project.scripts` must be a string')
 
                 sorted_scripts[name] = object_ref
 
@@ -902,9 +891,7 @@ class CoreMetadata(object):
 
             for name, object_ref in sorted(gui_scripts.items()):
                 if not isinstance(object_ref, str):
-                    raise TypeError(
-                        'Object reference `{}` of field `project.gui-scripts` must be a string'.format(name)
-                    )
+                    raise TypeError(f'Object reference `{name}` of field `project.gui-scripts` must be a string')
 
                 sorted_gui_scripts[name] = object_ref
 
@@ -934,23 +921,22 @@ class CoreMetadata(object):
             for forbidden_field in ('scripts', 'gui-scripts'):
                 if forbidden_field in defined_entry_point_groups:
                     raise ValueError(
-                        'Field `{0}` must be defined as `project.{0}` instead of in '
-                        'the `project.entry-points` table'.format(forbidden_field)
+                        f'Field `{forbidden_field}` must be defined as `project.{forbidden_field}` '
+                        f'instead of in the `project.entry-points` table'
                     )
 
             entry_point_groups = OrderedDict()
 
             for group, entry_point_data in sorted(defined_entry_point_groups.items()):
                 if not isinstance(entry_point_data, dict):
-                    raise TypeError('Field `project.entry-points.{}` must be a table'.format(group))
+                    raise TypeError(f'Field `project.entry-points.{group}` must be a table')
 
                 entry_points = OrderedDict()
 
                 for name, object_ref in sorted(entry_point_data.items()):
                     if not isinstance(object_ref, str):
                         raise TypeError(
-                            'Object reference `{}` of field `project.entry-points.{}` '
-                            'must be a string'.format(name, group)
+                            f'Object reference `{name}` of field `project.entry-points.{group}` must be a string'
                         )
 
                     entry_points[name] = object_ref
@@ -987,16 +973,16 @@ class CoreMetadata(object):
 
             for i, entry in enumerate(dependencies, 1):
                 if not isinstance(entry, str):
-                    raise TypeError('Dependency #{} of field `project.dependencies` must be a string'.format(i))
+                    raise TypeError(f'Dependency #{i} of field `project.dependencies` must be a string')
 
                 try:
                     requirement = Requirement(entry)
                 except InvalidRequirement as e:
-                    raise ValueError('Dependency #{} of field `project.dependencies` is invalid: {}'.format(i, e))
+                    raise ValueError(f'Dependency #{i} of field `project.dependencies` is invalid: {e}')
                 else:
                     if requirement.url and not self.hatch_metadata.allow_direct_references:
                         raise ValueError(
-                            'Dependency #{} of field `project.dependencies` cannot be a direct reference'.format(i)
+                            f'Dependency #{i} of field `project.dependencies` cannot be a direct reference'
                         )
 
                     dependencies_complex[get_normalized_dependency(requirement)] = requirement
@@ -1042,14 +1028,13 @@ class CoreMetadata(object):
             for option, dependencies in optional_dependencies.items():
                 if not is_valid_project_name(option):
                     raise ValueError(
-                        'Optional dependency group `{}` of field `project.optional-dependencies` must only contain '
-                        'ASCII letters/digits, underscores, hyphens, and periods, and must begin and end with '
-                        'ASCII letters/digits.'.format(option)
+                        f'Optional dependency group `{option}` of field `project.optional-dependencies` must only '
+                        f'contain ASCII letters/digits, underscores, hyphens, and periods, and must begin and end with '
+                        f'ASCII letters/digits.'
                     )
                 elif not isinstance(dependencies, list):
                     raise TypeError(
-                        'Dependencies for option `{}` of field `project.optional-dependencies` '
-                        'must be an array'.format(option)
+                        f'Dependencies for option `{option}` of field `project.optional-dependencies` must be an array'
                     )
 
                 entries = set()
@@ -1057,22 +1042,22 @@ class CoreMetadata(object):
                 for i, entry in enumerate(dependencies, 1):
                     if not isinstance(entry, str):
                         raise TypeError(
-                            'Dependency #{} of option `{}` of field `project.optional-dependencies` '
-                            'must be a string'.format(i, option)
+                            f'Dependency #{i} of option `{option}` of field `project.optional-dependencies` '
+                            f'must be a string'
                         )
 
                     try:
                         requirement = Requirement(entry)
                     except InvalidRequirement as e:
                         raise ValueError(
-                            'Dependency #{} of option `{}` of field `project.optional-dependencies` '
-                            'is invalid: {}'.format(i, option, e)
+                            f'Dependency #{i} of option `{option}` of field `project.optional-dependencies` '
+                            f'is invalid: {e}'
                         )
                     else:
                         if requirement.url and not self.hatch_metadata.allow_direct_references:
                             raise ValueError(
-                                'Dependency #{} of option `{}` of field `project.optional-dependencies` '
-                                'cannot be a direct reference'.format(i, option)
+                                f'Dependency #{i} of option `{option}` of field `project.optional-dependencies` '
+                                f'cannot be a direct reference'
                             )
 
                         entries.add(get_normalized_dependency(requirement))
@@ -1080,8 +1065,8 @@ class CoreMetadata(object):
                 normalized_option = normalize_project_name(option)
                 if normalized_option in normalized_options:
                     raise ValueError(
-                        'Optional dependency groups `{}` and `{}` of field `project.optional-dependencies` both '
-                        'evaluate to `{}`.'.format(normalized_options[normalized_option], option, normalized_option)
+                        f'Optional dependency groups `{normalized_options[normalized_option]}` and `{option}` of '
+                        f'field `project.optional-dependencies` both evaluate to `{normalized_option}`.'
                     )
 
                 normalized_options[normalized_option] = option
@@ -1103,7 +1088,7 @@ class CoreMetadata(object):
 
             for i, field in enumerate(dynamic, 1):
                 if not isinstance(field, str):
-                    raise TypeError('Field #{} of field `project.dynamic` must be a string'.format(i))
+                    raise TypeError(f'Field #{i} of field `project.dynamic` must be a string')
 
             self._dynamic = dynamic
 
@@ -1115,7 +1100,7 @@ class CoreMetadata(object):
             getattr(self, attribute)
 
 
-class HatchMetadata(object):
+class HatchMetadata:
     def __init__(self, root, config, plugin_manager):
         self.root = root
         self.config = config
@@ -1124,7 +1109,6 @@ class HatchMetadata(object):
         self._metadata = None
         self._build_config = None
         self._build_targets = None
-        self._version_source = None
         self._version = None
 
     @property
@@ -1163,18 +1147,6 @@ class HatchMetadata(object):
     @property
     def version(self):
         if self._version is None:
-            try:
-                self._version = self.version_source.get_version_data()['version']
-            except Exception as e:
-                raise type(e)(
-                    'Error getting the version from source `{}`: {}'.format(self.version_source.PLUGIN_NAME, e)
-                )  # TODO: from None
-
-        return self._version
-
-    @property
-    def version_source(self):
-        if self._version_source is None:
             if 'version' not in self.config:
                 raise ValueError('Missing `tool.hatch.version` configuration')
 
@@ -1182,9 +1154,37 @@ class HatchMetadata(object):
             if not isinstance(options, dict):
                 raise TypeError('Field `tool.hatch.version` must be a table')
 
-            options = deepcopy(options)
+            self._version = HatchVersionConfig(self.root, deepcopy(options), self.plugin_manager)
 
-            source = options.get('source', 'regex')
+        return self._version
+
+
+class HatchVersionConfig:
+    def __init__(self, root, config, plugin_manager):
+        self.root = root
+        self.config = config
+        self.plugin_manager = plugin_manager
+
+        self._cached = None
+        self._source_name = None
+        self._scheme_name = None
+        self._source = None
+        self._scheme = None
+
+    @property
+    def cached(self):
+        if self._cached is None:
+            try:
+                self._cached = self.source.get_version_data()['version']
+            except Exception as e:
+                raise type(e)(f'Error getting the version from source `{self.source.PLUGIN_NAME}`: {e}') from None
+
+        return self._cached
+
+    @property
+    def source_name(self):
+        if self._source_name is None:
+            source = self.config.get('source', 'regex')
             if not source:
                 raise ValueError(
                     'The `source` option under the `tool.hatch.version` table must not be empty if defined'
@@ -1192,16 +1192,59 @@ class HatchMetadata(object):
             elif not isinstance(source, str):
                 raise TypeError('Field `tool.hatch.version.source` must be a string')
 
-            version_source = self.plugin_manager.version_source.get(source)
+            self._source_name = source
+
+        return self._source_name
+
+    @property
+    def scheme_name(self):
+        if self._scheme_name is None:
+            scheme = self.config.get('scheme', 'standard')
+            if not scheme:
+                raise ValueError(
+                    'The `scheme` option under the `tool.hatch.version` table must not be empty if defined'
+                )
+            elif not isinstance(scheme, str):
+                raise TypeError('Field `tool.hatch.version.scheme` must be a string')
+
+            self._scheme_name = scheme
+
+        return self._scheme_name
+
+    @property
+    def source(self):
+        if self._source is None:
+            from copy import deepcopy
+
+            source_name = self.source_name
+            version_source = self.plugin_manager.version_source.get(source_name)
             if version_source is None:
-                raise ValueError('Unknown version source: {}'.format(source))
+                from ..plugin.exceptions import UnknownPluginError
 
-            self._version_source = version_source(self.root, options)
+                raise UnknownPluginError(f'Unknown version source: {source_name}')
 
-        return self._version_source
+            self._source = version_source(self.root, deepcopy(self.config))
+
+        return self._source
+
+    @property
+    def scheme(self):
+        if self._scheme is None:
+            from copy import deepcopy
+
+            scheme_name = self.scheme_name
+            version_scheme = self.plugin_manager.version_scheme.get(scheme_name)
+            if version_scheme is None:
+                from ..plugin.exceptions import UnknownPluginError
+
+                raise UnknownPluginError(f'Unknown version scheme: {scheme_name}')
+
+            self._scheme = version_scheme(self.root, deepcopy(self.config))
+
+        return self._scheme
 
 
-class HatchMetadataSettings(object):
+class HatchMetadataSettings:
     def __init__(self, root, config, plugin_manager):
         self.root = root
         self.config = config
@@ -1242,7 +1285,9 @@ class HatchMetadataSettings(object):
             for hook_name, config in hook_config.items():
                 metadata_hook = self.plugin_manager.metadata_hook.get(hook_name)
                 if metadata_hook is None:
-                    raise ValueError('Unknown metadata hook: {}'.format(hook_name))
+                    from ..plugin.exceptions import UnknownPluginError
+
+                    raise UnknownPluginError(f'Unknown metadata hook: {hook_name}')
 
                 configured_hooks[hook_name] = metadata_hook(self.root, config)
 
